# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Oct 10, 2017

@author: arnon
'''

from .. import pipe_listener_forever
import multiprocessing as mp
import threading as th
import logging 
import os 
from acrilib import LevelBasedFormatter, LoggerAddHostFilter
from .. import SSHPipeHandler
import signal
import sys

#from acrilib import LogSim

module_logger = logging.getLogger()

class SSHSubprocessClient(SSHPipeHandler):
    """ Pipe remote client wish message handler"""
        
    def __init__(self, cmd=[], calling_host=None, callback_port=None, *args, **kwargs):
        pass
        
def cmdargs():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default='127.0.0.1', required=False,
                        help='')
    parser.add_argument('--port', type=int, required=True,
                        help='')
    args = parser.parse_args()
    return vars(args)

if __name__ == '__main__':
    args = cmdargs()
    
    # TODO: add command line options
    
    client = SSHSubprocessClient(**args)
    client.service_loop()



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Oct 10, 2017

@author: arnon
'''

from ..pipe_listener import pipe_listener_forever
import multiprocessing as mp
import threading as th
import logging 
import os 
from acrilib import LevelBasedFormatter, LoggerAddHostFilter
from .. import SSHPipeHandler
import signal
import sys

#from acrilib import LogSim

module_logger = logging.getLogger()
        
def run(args, *, input=None, stdout=None, shell=False, cwd=None, timeout=None, check=False, encoding=None, errors=None):
    ''' Simulate python's subprocess run over sshpipe
    '''
    
    


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Aug 27, 2017

@author: arnon
'''

import pickle
import struct
import multiprocessing as mp
import subprocess as sp
import time
import logging
import os
from logging.handlers import QueueListener, QueueHandler
import signal
import enum
from acrilib import LoggerAddHostFilter


mlogger = logging.getLogger(__file__)


class SSHPipeQueueHandler(logging.Handler):
    def __init__(self, logger):
        super(SSHPipeQueueHandler, self).__init__()
        self.logger = logger

    def emit(self, record):
        self.logger.handle(record)


def stdbin_decode(value, encoding='utf8'):
    try:
        value = value.decode(encoding)
    except Exception:
        pass
    if value.endswith('\n'):
        value = value[:-1]
    return value


def remote_agent(where, command, pipe_read, pipe_write, communicateq, config,
                 encoding, name=None, callback=None, logq=None, ):
    ''' Runs SSH command overriding its stdin and stdout with pipes.

    Args:
        where: user@host or host string.
        command: string or list of command parts to run under ssh. 
        pipe_read, pipe_write: read and write ends of multiprocessing Pipe().
        communicateq: multiprocessing queue to puss back response.
        callback: to invoke when agent is terminated.

    remote_agent is intended to run as a mulitprocessing process.  It uses
    subprocess run to run ssh command on remote host. remote_agent overrides
    subprocess stdin of subprocess with pipe_read.

    remote_agent collects stdout and stderr and pass them back via
    communicate queue.
    '''
    global mlogger

    if logq is not None:
        # if name is not None:
        mlogger = logging.getLogger()
        mlogger.addHandler(QueueHandler(logq))
        mlogger.addFilter(LoggerAddHostFilter())
    pipe_write.close()

    # convert pipe end to opened file descriptor, so it can be passed to
    # subprocess.
    pipe_readf = os.fdopen(os.dup(pipe_read.fileno()), 'rb')

    cmd = ["ssh"]
    if config:
        cmd.append("-F {}".format(config))

    if isinstance(command, str):
        command = [command]
    cmd += [where] + command

    mlogger.debug('Starting subprocess run({})'.format(cmd))
    sshrun = sp.run(cmd, shell=False, stdin=pipe_readf, stdout=sp.PIPE,
                    stderr=sp.PIPE, check=False,)

    returncode, stdout, stderr = (sshrun.returncode,
                                  sshrun.stdout.decode(encoding),
                                  sshrun.stderr.decode(encoding))
    response = (returncode, stdout, stderr)

    error = '' if not stderr else '\n' + stderr
    mlogger.debug(('Response placed in SSH queue: returncode: {}, '
                   'stdout: {}, stderr: {}').format(repr(returncode),
                                                    stdout, error))
    communicateq.put(response)
    if callback is not None:
        callback(returncode, stdout, stderr)
    pipe_readf.close()
    exit(returncode)


class SSHPipeState(enum.Enum):
    initiated = 0
    started = 1
    closed = 2


class SSHPipe(object):
    ''' Facilitates creation of agent process on remote host using ssh and
    feeding that process with data to process.
    '''
    def __init__(self, host, command, name=None, user=None,
                 term_message='TERM', config=None, encoding='utf8',
                 callback=None, logger=None):
        ''' SSHPipe object runs command on remote host with the intention to
        pass workloads throughout the lifetime of the run.

        Assumptions:
            ssh key configuration among cluster hosts.
                set .profile with: source /var/venv/eventor/bin/activate
                create environment ssh keys:
                    ssh-keygen -t rsa -C "envname" -f "id_rsa_envname"
                add to authorized_keys proper key:

command=". ~/.profile_envname; if [ -n \"$SSH_ORIGINAL_COMMAND\" ];
then eval \"$SSH_ORIGINAL_COMMAND\"; else exec \"$SHELL\"; fi" ssh-rsa ...

        Args:
            host: remote host.
            command: to run on remote host. String or list of command string parts.
            user: user to use to access remote host.
            config: SSH configuration to use. ~/.ssh/config .
            callback: if available, would be invoked once the agent in remote
            host is terminated.
                e.g., callback maybe a method that place result in a queue or
                send to a port. callback function should have the following
                calling arguments:returncode, stdout, stderr
            encoding: to use in decoding stdout and stderr.
            logger: logger object (logging).
        '''
        global mlogger
        if logger is not None:
            mlogger = logger

        self.term_message = term_message

        self.pipe_read, self.pipe_write = mp.Pipe()
        self.__qlogger = None
        self.__logq = mp.Queue()

        self.__communicateq = mp.Queue()
        self.where = \
            "{}{}".format('' if user is None else "@{}".format(user), host)
        self.command = command
        self.callback = callback

        self.pipe_writef = None
        self.result = None
        self.config = config
        self.encoding = encoding
        self.name = name
        self.__state = SSHPipeState.initiated

    def start(self, wait=1):
        ''' Starts process wherein SSH command will be executed.

        Args:
            wait: seconds to sleep between check process status.
        '''
        global mlogger

        if self.__state == SSHPipeState.started:
            return

        self.__qlogger = QueueListener(self.__logq,
                                       SSHPipeQueueHandler(mlogger))
        self.__qlogger.start()

        remote_agent_kwargs = {
            'where': self.where,
            'command': self.command,
            'pipe_read': self.pipe_read,
            'pipe_write': self.pipe_write,
            'communicateq': self.__communicateq,
            'config': self.config,
            'encoding': self.encoding,
            'name': self.name,
            'logq': self.__logq,
            'callback': self.callback,
            # 'logger':mlogger,
            }

        try:
            self.__agent = mp.Process(target=remote_agent,
                                      kwargs=remote_agent_kwargs, daemon=True)
            self.__agent.start()
        except Exception:
            raise
        self.pid = self.__agent.pid

        mlogger.debug("Agent activated: {}.".format(self.__agent.pid))

        if wait is not None:
            while True:
                time.sleep(wait)
                mlogger.debug("Waiting for start: alive: {}, exitcode: {}"
                              .format(self.__agent.is_alive(),
                                      self.__agent.exitcode))
                if self.__agent.is_alive() or \
                   self.__agent.exitcode is not None:
                    break

        if self.is_alive():
            mlogger.debug("Agent is running: {}.".format(self.__agent.pid))
            self.__state = SSHPipeState.started
        else:
            mlogger.debug("Agent is dead: {}.".format(self.response()))
            self.close()

        self.pipe_read.close()
        self.pipe_writef = os.fdopen(os.dup(self.pipe_write.fileno()), 'wb')

        signal.signal(signal.SIGHUP, self.exit_gracefully)
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def is_alive(self):
        ''' Returns True if remote_agent process is alive.
        '''
        return self.__agent.is_alive() and self.__communicateq.empty()

    def prepare(self, msg, pack=True):
        ''' Returns byte message prefixed by its length packed into bytes.

        Args:
            msg: object to prepate.
            pack: if set, pickle dumps msg.
        '''
        workload = msg
        if pack:
            workload = pickle.dumps(msg)
        msgsize = len(workload)
        mlogger.debug('Preparing message of size: %s.' % (msgsize,))
        msgsize_packed = struct.pack(">L", msgsize)
        return msgsize_packed + workload

    def send(self, msg, pack=True):
        ''' Writes msg into remote_agent pipe (process' stdin)

        Args:
            msg: message object
            pack: if set, pickle dumps msg.
        '''
        request = self.prepare(msg, pack=pack)
        mlogger.debug(
             'Writing message of actual size {} to pipe.'
             .format(len(request)))

        self.pipe_writef.write(request)

    def response(self, timeout=None):
        ''' Wait for response from remote_agent.

        Args:
            timeout: if set, limits wait time.

        Returns:
            (returncode, stdout, stderr) resulted by remote_agent
        '''
        global mlogger

        mlogger.debug('Getting response from SSH control queue.')
        if self.result is None:
            try:
                result = self.__communicateq.get(timeout=timeout)
            except Exception:
                result = None

            if result:
                encoding = self.encoding
                returncode, stdout, stderr = (
                    result[0], stdbin_decode(result[1], encoding=encoding),
                    stdbin_decode(result[2], encoding=encoding))
                if not stdout:
                    stdout = self.term_message
                self.result = returncode, stdout, stderr
        mlogger.debug('Received from SSH control queue: {}'
                      .format(repr(self.result)))
        return self.result

    def close(self, msg=None):
        ''' Sends termination message to remote_agent.

        Args:
            msg: termination message, defaults to 'TERM'

        Returns:
            same as response()
        '''
        global mlogger

        if self.__state == SSHPipeState.closed:
            return self.response()

        if msg is None:
            msg = self.term_message

        if self.is_alive():
            mlogger.debug('Sending {} to pipe.'.format(msg))
            self.send(msg)
        else:
            mlogger.debug('Process is not alive, skipping {}.'.format(msg))

        if self.pipe_writef:
            try:
                self.pipe_writef.close()
            except BrokenPipeError:
                # already closed
                mlogger.debug('BorkenPipe when closing writer; Reader already closed.')

        response = self.response()
        mlogger.debug('Joining with subprocess.')
        self.__agent.join()
        mlogger.debug('Subprocess joined.')
        if self.__qlogger is not None:
            self.__qlogger.stop()
        self.__state = SSHPipeState.closed
        return response

    def exit_gracefully(self, signo, stack_frame, *args, **kwargs):
        self.close()

    def __del__(self):
        self.close()

    def join(self, timeout=None):
        ''' Waits for remote_agent to finish.

        Args:
            timeout: limits time wait as in Process.join()
        '''
        self.__agent.join(timeout)



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import sys
import struct
import pickle
import logging
import importlib
import os
from acrilib import TimedSizedRotatingHandler
# from acrilog.mplogger import MpLogger

module_logger = logging.getLogger(__name__)

EXIT_MESSAGE = 'TERM'


def set_logger(name=None, logdir='/tmp'):
    if name is not None:
        logger_file = name
    else:
        logger_file = "{}.{}.log"\
            .format(os.path.basename(__file__).rpartition('.')[0], os.getpid())
    logger_file = os.path.join(logdir, logger_file)
    logger_handler = TimedSizedRotatingHandler(logger_file,)
    logger = logging.getLogger()
    logger.addHandler(logger_handler)
    return logger


def imports_from_cmd(imports_str):
    imports = dict()

    for import_str in imports_str:
        import_file, _, import_modules_str = import_str.partition(':')
        import_modules = import_modules_str.split(':')
        file_modules = imports.get(import_file, list())
        file_modules.extend(import_modules)
        imports[import_file] = file_modules
    imports = [(import_file, set(modules))
               for import_file, modules in imports.items() if modules]
    return imports


def _imports(imports, queue, exit_message):
    imports = imports_from_cmd(imports)
    module_logger.debug("Importing {}.".format(imports))
    for import_file, import_modules in imports:
        if not import_file:
            for module in import_modules:
                module_logger.debug("Importing %s." % (module))
                try:
                    from importlib import import_module
                    import_module(module)
                except Exception as e:
                    module_logger.critical("Failed to import: %s." % (module))
                    module_logger.exception(e)
                    # signal to parent via stdout
                    print('TERM')
                    print(e, file=sys.stderr)
                    queue.put((exit_message, e))
                    return False
        else:
            for module in import_modules:
                module_logger.debug("Importing %s from %s."
                                    .format(module, import_file))
                try:
                    spec = importlib.util.spec_from_file_location(module, import_file)
                    spec_module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(spec_module)
                    # sys.modules[args.import_module] = module # not needed for now
                except Exception as e:
                    module_logger.critical("Failed to import: %s %s;"
                                           .format(import_module, import_file))
                    module_logger.exception(e)
                    # signal to parant via stdout
                    print('TERM')
                    print(e, file=sys.stderr)
                    queue.put((exit_message, e))
                    return False
    return True


def pipe_recv_message(queue, exit_message=EXIT_MESSAGE):
    try:
        msgsize_raw = sys.stdin.buffer.read(4)
    except Exception as e:
        module_logger.critical('Failed to read STDIN.')
        module_logger.exception(e)
        queue.put((exit_message, e))
        return

    try:
        msgsize = struct.unpack(">L", msgsize_raw)
    except Exception as e:
        module_logger.critical('Failed pickle loads message size from STDIN; received: %s' % hex(msgsize_raw))
        module_logger.exception(e)
        queue.put((exit_message, e))
        return

    try:
        msg_pack = sys.stdin.buffer.read(msgsize[0])
    except Exception as e:
        module_logger.critical('Failed to read STDIN.')
        module_logger.exception(e)
        queue.put((exit_message, e))
        return


def pipe_listener(queue, term_messages=[EXIT_MESSAGE], unpack=False):
    ''' Pipe listener wait for one message.  Once receive (DONE or TERM) it ends.

    Args:
        imports: string in the form of [file]:module[:module ...]
            if file, all modules will be imported from file.
    '''
    global module_logger

    # in this case, whiting for possible termination message from server
    exit_message = term_messages[0]

    module_logger.debug('Trying to read STDIN message size.')
    try:
        msgsize_raw = sys.stdin.buffer.read(4)
    except Exception as e:
        module_logger.critical('Failed to read STDIN.')
        module_logger.exception(e)
        queue.put((exit_message, e))
        return False

    module_logger.debug('Trying to unpack message size: {}.'
                        .format(repr(msgsize_raw)))
    try:
        msgsize = struct.unpack(">L", msgsize_raw)
    except Exception as e:
        module_logger.critical(
            'Failed pickle loads message size from STDIN; received: {}.'
            .format(hex(msgsize_raw)))
        module_logger.exception(e)
        queue.put((exit_message, e))
        return False

    msgsize = msgsize[0]
    module_logger.debug('Trying to read STDIN message of size {}.'
                        .format(msgsize))
    try:
        msg_pack = sys.stdin.buffer.read(msgsize)
    except Exception as e:
        module_logger.critical('Failed to read STDIN.')
        module_logger.exception(e)
        queue.put((exit_message, e))
        return False

    if unpack:
        try:
            msg = pickle.loads(msg_pack)
        except Exception as e:
            module_logger.critical('Failed pickle loads message from STDIN.')
            module_logger.exception(e)
            queue.put((exit_message, e))
            return False
    else:
        msg = msg_pack

    module_logger.debug(
        'Received message from remote parent: {}; passing to main process.'
        .format(msg))

    queue.put((msg, ''))

    return msg not in term_messages  # == exit_message


def pipe_listener_forever(message_queue, started_event,
                          term_messages=[EXIT_MESSAGE], unpack=True,
                          imports=None, logger=None):
    ''' Same us pipe_listener, but loop until request to abort
    '''
    global module_logger

    if logger is not None:
        # module_logger = MpLogger.get_logger(logger_info)
        module_logger = logger

    if imports is not None:
        imported = _imports(imports, message_queue, term_messages[0])
        if not imported:
            return
    started_event.set()

    active = True
    while active:
        active = pipe_listener(message_queue, term_messages=term_messages,
                               unpack=unpack)



'''
Created on Oct 19, 2017

@author: arnon
'''
import sys
import struct
import pickle


class SSHPipeCallableHandler(object):
    def __init__(self, term_message='TERM',):
        self.term_message = term_message

    def service_loop(self):
        while True:
            try:
                msgsize_raw = sys.stdin.buffer.read(4)
                msgsize = struct.unpack(">L", msgsize_raw)
                workload = sys.stdin.buffer.read(msgsize[0])
                worker = pickle.loads(workload)
            except Exception as e:
                print(e, file=sys.stderr)
                print(self.term_action)
                exit(1)

            retcode = self.action(worker)
            if worker is not None:
                exit(retcode)

    def action(self, received):
        if not isinstance(received, str):
            received()
            return None
        elif received == self.term_message:
            # maybe worker prints to stdout
            return 0
        else:
            print("Bad object received: " + repr(received), file=sys.stderr)
            print(self.term_action)
            return 2



#!/usr/bin/env python3
'''
Created on Oct 18, 2017

@author: arnon
'''

from sshpipe import SSHPipeHandler
import logging
from logging.handlers import SocketHandler
import pickle
import struct

mlogger = logging.getLogger(__file__)


class SSHPipeSocketHandler(SSHPipeHandler, SocketHandler):
    ''' SSHPipeSocketHandler modeled over logging.handlers.SocketHandler
    '''

    def __init__(self, port, host='17.0.0.1', *args, **kwargs):
        SSHPipeHandler.__init__(self, *args, **kwargs)
        SocketHandler.__init__(self, host, port)

        if port is None:
            self.address = host
        else:
            self.address = (host, port)
        self.mlogger.debug("Server socket address: {}.".format(self.address))

    def makePickle(self, record):
        """
        Pickles the record in binary format with a length prefix, and
        returns it ready for transmission across the socket.
        """
        if isinstance(record, str):
            s = pickle.dumps(record)
        else:  # object
            d = dict(record.__dict__)
            s = pickle.dumps(d, 1)
        slen = struct.pack(">L", len(s))
        return slen + s

    def atstart(self, receieved):
        # file = "{}{}".format(__file__, ".remote.log")
        # self.mlogger.debug("Opening file: {}.".format(file))
        # self.file = open(file, 'w')

        # call create socket to prevent it being called at first handle.
        self.createSocket()

    def atexit(self, received):
        # if self.file is not None:
        #     self.file.close()
        SSHPipeHandler.atexit(self, received)
        SocketHandler.close(self)
        self.mlogger.debug("Closed handlers.")

    def handle(self, received):
        """
        Send a pickled string to the socket.

        This function allows for partial sends which can happen when the
        network is busy.
        """
        # self.file.write(str(received))
        self.emit(received)
        self.mlogger.debug("Emitted record to socket: {}.".format(received))


def cmdargs():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default='127.0.0.1', required=False, help='')
    parser.add_argument('--port', type=int, required=False, help='')
    parser.add_argument("--id", type=str, required=False, default=1, dest='handler_id')
    args = parser.parse_args()
    return vars(args)


if __name__ == '__main__':
    args = cmdargs()

    # TODO: add command line options

    client = SSHPipeSocketHandler(**args)
    client.service_loop()



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Oct 3, 2017

@author: arnon
'''
import os
from sshpipe import SSHPipe

'''
    ssh_tunnel.py instantiate an SSH pipe on remote host.
    At the far end of the pipe, it runs a Pipe Listener server that would
    receive messages and act upon.
'''


class SSHTunnel(object):
    ''' Creates and operates SSH tunnel to remote host.

    SSH Tunnel is based on SSHPipe starting in one end and SSHPipeHandler
    based executable running in the remote end.

    '''

    def __init__(self, remote, receiver, name=None, logger=None):
        ''' Initiates SSHTunnle object.

        Args:
            remote: ssh key tag configured in .ssh/config such that is set
                in target host to start with the right environment.
            receiver: executable that establishes SSHPipeHandler object.
                String or list of command string parts.

        '''
        self.receiver = receiver
        self.remote = remote
        self.__sshagent = None
        self.__state = 'initial'
        self.logger = logger
        self.name = name

    def start(self, wait=1):
        self.__sshagent =\
            sshagent = SSHPipe(host=self.remote, command=self.receiver, name=self.name, logger=self.logger)
        sshagent.start(wait=wait)

        if not sshagent.is_alive():
            raise RuntimeError("SSHAgent is not alive.")

        self.__state = 'start'
        return self

    def pid(self):
        return self.__sshagent.pid

    def send(self, message, pack=True):
        if self.__state != 'start':
            raise RuntimeError("Attempted send() without starting the tunnel.")
        self.__sshagent.send(message, pack=pack)

    def is_alive(self):
        return self.__sshagent.is_alive()

    def response(self):
        if self.__state != 'close':
            raise RuntimeError("Attempted response() without closing the tunnel.")
        return self.__sshagent.response()

    def close(self, msg=None):
        if self.__state == 'start':
            self.__sshagent.close(msg=msg)
            self.__state = 'close'



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Oct 10, 2017

@author: arnon
'''

from .pipe_listener import pipe_listener_forever
import multiprocessing as mp
import threading as th
import logging
from logging.handlers import RotatingFileHandler, QueueHandler, QueueListener
import os
from acrilib import LevelBasedFormatter, LoggerAddHostFilter
import signal
import sys
from sshpipe import SSHPipe

mlogger = logging.getLogger()


class SSHPipeHandler(object):
    """ Pipe remote client wish message handler"""

    logging_config = {
        'level_formats': {
            logging.DEBUG: ("[ %(asctime)-15s ][ %(host)s ]"
                            "[ %(processName)-11s ][ %(levelname)-7s ]"
                            "[ %(message)s ]"
                            "[ %(module)s.%(funcName)s(%(lineno)d) ]"),
            'default': ("[ %(asctime)-15s ][ %(host)s ]"
                        "[ %(processName)-11s ][ %(levelname)-7s ]"
                        "[ %(message)s ]"),
            },
        'level': logging.DEBUG,
        'logdir': '/tmp',
        'encoding': 'utf8',
        'mode': 'a',
        'maxBytes': 0,
        'backupCount': 0,
        'delay': False,
        }

    TERM_MESSAGES = ['TERM', 'STOP', 'FINISH', 'EXIT', 'QUIT']

    def __init__(self, handler_id=None, term_messages=None, unpack=True,
                 imports=None, tolerance=0, caller_host=None,
                 caller_port=None):
        """
        Args:
            tolerance: indicate how to behave if couldn't understand message
                or handling failed.
                -1: tolerate all exception in message reception or handling.
                 0: zero tolerance, abort on first error
                 n: allow n errors before aborting (abort on error number n+1)
        """
        global mlogger
        self.term_messages = term_messages
        if term_messages is None:
            self.term_messages = SSHPipeHandler.TERM_MESSAGES
        self.term_message = self.term_messages[0]
        self.unpack = unpack
        self.imports = imports
        self.tolerance = tolerance
        self.handler_id = handler_id
        if handler_id is None:
            self.handler_id = "sshpipe_handler_{}".format(os.getpid())
        self._exitmsg = ""
        self._exitcode = 0
        self.caller_host = caller_host
        self.caller_port = caller_port
        self._pipe = None

        sshpipe_config_file = \
            os.environ.get("SSHPIPE_CONFIG_FILE",
                           os.path.expanduser('~/.sshpipe'))
        file_config = {}
        if os.path.isfile(sshpipe_config_file):
            # TODO: read configuration for logging
            #       (logdir, formats, level, etc.)
            pass
        config = SSHPipeHandler.logging_config
        config.update(file_config)

        logq = mp.Queue()
        logfile = "{}.log".format(self.handler_id)
        logdir = config['logdir']
        if logdir is not None:
            logfile = os.path.join(logdir, logfile)

        file_handler_kwargs = {
            'filename': logfile,
            'mode': config['mode'],
            'maxBytes': config['maxBytes'],
            'backupCount': config['backupCount'],
            'encoding': config['encoding'],
            'delay': config['delay'],
            }

        level = config['level']
        formatter = LevelBasedFormatter(config['level_formats'])
        handler = RotatingFileHandler(**file_handler_kwargs)
        handler.setFormatter(formatter)
        handler.setLevel(level)
        self.qlogger = QueueListener(logq, handler, respect_handler_level=False)
        self.qlogger.start()

        queue_handler = QueueHandler(logq)
        mlogger.addHandler(queue_handler)
        mlogger.setLevel(level)
        mlogger.addFilter(LoggerAddHostFilter())
        self.mlogger = mlogger
        mlogger.debug("self.mlogger: method is set {}.".format(repr(self.mlogger)))

        signal.signal(signal.SIGHUP, self.exit_gracefully)
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def message_queue_loop(self, ):
        ''' loops on message_queue running handler on messages received.
        If control message receive, run appropriate control.

        message_queue_loop is design to run on a parallel thread to loop
        forever.
        Pipe listener's loop_forever fetch messages from STDIN into message queue for
        message_queue_loop to handle.
        '''
        global mlogger
        visited_atstart = False
        while True:
            mlogger.debug("Getting from message_queue.")
            received = self.message_queue.get()

            if received is None:
                continue

            message, error = received
            errmsg = '' if not error else "; error: {}.".format(error)
            mlogger.debug("Message queue received: message: {}{}"
                          .format(repr(message), errmsg))

            if error:
                if self.tolerance == 0:
                    self._exitmsg = error
                    self._exitcode = 1
                    break
                self.tolerance -= 1

            if not visited_atstart:
                if self.caller_host and self.caller_port:
                    self._create_callback_channel()
                    self._send_callback('ATSTART {}'.format(self.handler_id))

                mlogger.debug("Message queue not visited atstart yet.")
                visited_atstart = True
                try:
                    self.atstart(message)
                except Exception as error:
                    mlogger.exception("Exception from atstart: {}."
                                      .format(error))
                    self._exitmsg = error
                    self._exitcode = 2
                    break

            if message in self.term_messages:
                break

            try:
                self.handle(message)
            except Exception as error:
                mlogger.exception("Exception from handler: {}."
                                  .format(error))
                self._exitmsg = error
                self._exitcode = 3
                break

        self._send_callback('ATEXIT {}'.format(self.handler_id))
        self._close_callback_channel()

        try:
            self.atexit(message)
        except Exception as error:
            mlogger.exception("Exception from atexit: {}.".format(error))
            self._exitmsg = error
            self._exitcode = 4
        loop_message = "." if self._exitcode == 0 else " with error."
        msg = "Completed message queue loop{}.".format(loop_message)
        mlogger.debug(msg)

    def service_loop(self):
        global mlogger
        self.message_queue = mp.Queue()
        self.started_event = mp.Event()

        pipe_listener_forever_kwargs = {
            'message_queue': self.message_queue,
            'started_event': self.started_event,
            'term_messages': self.term_messages,
            'unpack': self.unpack,
            'imports': self.imports,
            'logger': mlogger,
            }

        message_queue_loop_th = th.Thread(target=self.message_queue_loop,
                                          daemon=True)
        message_queue_loop_th.start()

        pipe_listener_forever(**pipe_listener_forever_kwargs)

        message_queue_loop_th.join()

        mlogger.debug("Completed service loop; exitcode: {}; exitmsg:{}."
                            .format(self._exitcode, self._exitmsg))

        if self._exitcode > 0:
            dest = sys.stderr
            print(self._exitmsg, dest)
            # TODO: exit must be a parameter choice
            exit(self._exitcode)

    def _create_callback_channel(self):
        global mlogger
        command = "sshpipe_socket_handler.py --port {}".format(
            self.caller_port)
        self._pipe = SSHPipe(host=self.caller_host, command=command,
                             logger=mlogger)

    def _send_callback(self, msg):
        if self._pipe:
            self._pipe.send(msg)

    def _close_callback_channel(self):
        if self._pipe:
            response = self.pipe.close()
            return response

    def atstart(self, received):
        global mlogger
        mlogger.debug("atstart: nothing to do, message: {}."
                      .format(received))

    def handle(self, received):
        global mlogger
        mlogger.debug("handle: nothing to do, message: {}."
                      .format(received))

    def atexit(self, received):
        global mlogger
        mlogger.debug("atexit: with message {}".format(received))
        self.qlogger.stop()
        msg = received
        if received not in self.term_messages:
            msg = self.term_message
        print(msg)

    def exit_gracefully(self, signo, stack_frame, *args, **kwargs):
        self.atexit(self.term_message)
        print(self.term_message)



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
from sshpipe.lib.sshconfig import load, loads, dump, dumps
from sshpipe.lib.sshpipe import SSHPipe
from sshpipe.lib.sshpipe_callable_handler import SSHPipeCallableHandler
from sshpipe.lib.pipe_listener import pipe_listener, pipe_listener_forever, EXIT_MESSAGE, set_logger
from sshpipe.lib.sshpipe_handler import SSHPipeHandler
from sshpipe.lib.ssh_tunnel import SSHTunnel
# from sshpipe.lib.ssh_chain import SSHChain
# from sshpipe.lib.ssh_chain_handler import SSHChainHandler

__author__ = 'arnon'
__version__ = '0.5.2'




import sys

from setuptools import setup
import importlib.util

''' 
is the Python package in your project. It's the top-level folder containing the 
__init__.py module that should be in the same directory as your setup.py file
/-
  |- README.rst
  |- CHANGES.txt
  |- setup.py
  |- dogs 
     |- __init__.py
     |- catcher.py

To create package and upload:

  python setup.py register
  python setup.py sdist
  twine upload -s dist/package.tar.gz

'''


def import_setup_utils():
    # load setup utils
    try:
        setup_utils_spec = \
            importlib.util.spec_from_file_location("setup.utils",
                                                   "setup_utils.py")
        setup_utils = importlib.util.module_from_spec(setup_utils_spec)
        setup_utils_spec.loader.exec_module(setup_utils)
    except Exception as err:
        raise RuntimeError("Failed to find setup_utils.py."
                           " Please copy or link.") from err
    return setup_utils


setup_utils = import_setup_utils()
PACKAGE = "sshpipe"
NAME = PACKAGE
DESCRIPTION = ('sshpipe provide tools to manage ssh channel to remote hosts.')

metahost = setup_utils.metahost(PACKAGE)
AUTHORS, AUTHOR_EMAILS = setup_utils.read_authors(metahost=metahost)
URL = 'https://github.com/Acrisel/sshpipe'
VERSION = setup_utils.read_version(metahost=metahost)

# Warn if we are installing over top of an existing installation. This can
# cause issues where files that were deleted from a more recent Accord are
# still present in site-packages. See #18115.
overlay_warning = False
if "install" in sys.argv:
    existing_path = setup_utils.existing_package(PACKAGE)

scripts = setup_utils.scripts(PACKAGE)

# Find all sub packages
packages = setup_utils.packages(PACKAGE)
required = setup_utils.read_required(metahost=metahost)

setup_info={
    'name': NAME,
    'version': VERSION,
    'url': URL,
    'author': AUTHORS,
    'author_email': AUTHOR_EMAILS,
    'description': DESCRIPTION,
    'long_description': open("README.rst", "r").read(),
    'license': 'MIT',
    'keywords': 'library logger multiprocessing',
    'packages': packages,
    'scripts': scripts,
    'install_requires': required,
    'extras_require': {'dev': [], 'test': []},
    'classifiers': ['Development Status :: 4 - Beta',
                    'Environment :: Other Environment',
                    #'Framework :: Project Settings and Operation',
                    'Intended Audience :: Developers',
                    'License :: OSI Approved :: MIT License',
                    'Operating System :: OS Independent',
                    'Programming Language :: Python',
                    'Programming Language :: Python :: 3',
                    'Topic :: System :: Distributed Computing',
                    'Topic :: Software Development :: Libraries :: Python '
                    'Modules']}
setup(**setup_info)


if overlay_warning:
    sys.stderr.write("""

========
WARNING!
========

You have just installed %(name)s over top of an existing
installation, without removing it first. Because of this,
your install may now include extraneous files from a
previous version that have since been removed from
Accord. This is known to cause a variety of problems. You
should manually remove the

%(existing_path)s

directory and re-install %(name)s.

""" % {"existing_path": existing_path, 'name': NAME})



#!/usr/bin/env python

# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import logging
from sshpipe.sshpipe import SSHPipe
from sshpipe.pipe_listener import pipe_listener_forever, EXIT_MESSAGE
import multiprocessing as mp

module_logger = logging.getLogger(__name__)


def remote_socket_agent(port, exit_message=EXIT_MESSAGE, pickle_loads=False, imports=None):
    ''' Forward SSH messages to port
    '''
    global module_logger
    controlq = mp.Queue()
    kwargs = {'queue':controlq, 'port':port, 'exit_message':exit_message, 'pickle_loads': pickle_loads, 'imports':imports, 'logger': module_logger,}
    listener = mp.Process(target=pipe_listener_forever, kwargs=kwargs, daemon=True)
    listener.start()
    
    active = True
    while active:
        msg = controlq.get()
        active = msg == EXIT_MESSAGE
        
        
class SSHRemotePort(SSHPipe):
    ''' Facilitates sending information to remote host:port via ssh channel
    '''
    def __init__(self, host, port, pack=False, exit_message=None, imports=None, logger=None, *args, **kwargs):
        self.host = host
        self.port = port
        self.pack = pack
        cmd_parts= ["ssh_remote_port.py --port {}".format(port)]
        if imports:
            cmd_parts.append("--imports {}".format(imports))
        if pack:
            cmd_parts.append("--pickle-loads")
        if exit_message:
            cmd_parts.append("--exit-message {}".format(exit_message))
            
        command = ' '.join(cmd_parts)
        super(SSHRemotePort, self).__init__(self.host, command, *args, logger=logger, **kwargs) #=module_logger)
        
    def start(self):
        super(SSHRemotePort, self).start(wait=0.2)
        

def cmdargs():
    import argparse
    import os
    
    filename = os.path.basename(__file__)
    progname = filename.rpartition('.')[0]
    
    parser = argparse.ArgumentParser(description="%s runs SSH Port Agent" % progname)
    parser.add_argument('--port', type=int, 
                        help="""Port to forward messages to.""")
    parser.add_argument('--exit-message', type=str, dest='exit_message', default=EXIT_MESSAGE,
                        help="""string to use as exit message, default: {}.""".format(EXIT_MESSAGE))
    parser.add_argument('--pickle-loads', action='store_true', default=False, dest='pickle_loads',
                        help='if set, will pikcle.loads() message before sending to port.')
    parser.add_argument('--imports', type=str, required=False, dest='imports', nargs='*',
                        help="""import module before pickle loads.""")
    args = parser.parse_args()  
    
    return args
    
if __name__ == '__main__':
    mp.freeze_support()
    mp.set_start_method('spawn')
    
    args = cmdargs()
    remote_socket_agent(args)
    
    